#ifndef __PDE_KERNELS_CU
#define __PDE_KERNELS_CU

#include <stdio.h>
#include <string.h>

//#define DEBUG_KERNEL

#define ATD(x,y) (pitch)*(gridPosBase_j+(y)) + (gridPosBase_i+(x))
#define ATS(x,y) (local_w)*(y) + (x)

texture<float, 1, cudaReadModeElementType> texRef_0;
texture<float, 1, cudaReadModeElementType> texRef_1;
texture<float, 1, cudaReadModeElementType> texRef_2;
texture<float, 1, cudaReadModeElementType> texRef_3;

float __device__ selectTex1D(int gpu, int pos)
{
	switch(gpu)
	{
		case 0:
			return tex1Dfetch(texRef_0, pos);
		case 1:
			return tex1Dfetch(texRef_1, pos);
		case 2:
			return tex1Dfetch(texRef_2, pos);
		case 3:
			return tex1Dfetch(texRef_3, pos);
	}
	
	return 0.f;
}

/*
 * Red SOR through GlobMem r/w.
 */

__global__ void rsor_gm(float *dgrid, int n, int h, float omega, int pitch, int abs_start, int dev)
{
	int gridPosBase_j = blockIdx.y*2; //window size = 2
	int gridPosBase_i = blockIdx.x*(blockDim.x-2);
	int i = threadIdx.x;
	int local_w = blockDim.x;
	
	extern __shared__ float sgrid[];
  
  //If block can't use more than 3 rows
  int ok3 = gridPosBase_j+3 < h;
	//Red first
  int r1 = (abs_start+gridPosBase_j +1)&1;
  //Traversing sign 
  int sign = r1 + __mul24(1-r1, -1);
	//Computation indexes
	int com_i = i+1;
	int com_j = (2-r1) + __mul24(sign, 1&i);
#ifdef DEBUG_KERNEL
	if(i==0)
		printf("t%d: RED abs_pos: %d\n", dev, abs_start + gridPosBase_j +1);
#endif
	
	if(gridPosBase_i+i < n)
	{
	
#ifdef DEBUG_KERNEL
		if(i==0)
		{
			printf("t%d: RED grid pos: (%d,%d)\n", dev, blockIdx.x, blockIdx.y);
			printf("t%d: RED ok3: %d, r1: %d, sign: %d\n", dev, ok3, r1, sign);
		}
#endif
		
		sgrid[ATS(i,0)] = dgrid[ATD(i,0)];
		sgrid[ATS(i,1)] = dgrid[ATD(i,1)];
		sgrid[ATS(i,2)] = dgrid[ATD(i,2)];
		if(ok3)
			sgrid[ATS(i,3)] = dgrid[ATD(i,3)];
	}
  
	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("RED pre=============================================\n");
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif
		
	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
		else if(com_j != 2)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
	}

	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("RED post=============================================\n");
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			dgrid[ATD(com_i, com_j)] = sgrid[ATS(com_i, com_j)];
		else if(com_j == 1)
			dgrid[ATD(com_i, 1)] = sgrid[ATS(com_i, 1)];
	}

}

/*
 * Black SOR through GlobMem r/w.
 */

__global__ void bsor_gm(float *dgrid, int n, int h, float omega, int pitch, int abs_start, int dev)
{
	int gridPosBase_j = blockIdx.y*2; //window size = 2
	int gridPosBase_i = blockIdx.x*(blockDim.x-2);
	int i = threadIdx.x;
	int local_w = blockDim.x;
	extern __shared__ float sgrid[];
  
  int ok3 = gridPosBase_j+3 < h;
	//Red first
  int r1 = (abs_start + gridPosBase_j +1)&1;
  //Traversing sign 
  int sign = __mul24(-1,r1)+(1-r1);
	//Computation indexes
	int com_i = i+1;
	int com_j = (1+r1) + __mul24(sign,1&i);

#ifdef DEBUG_KERNEL
	if(i==0)
		printf("t%d: BLACK abs_pos: %d\n", dev, abs_start + gridPosBase_j +1);
#endif
  
	if(gridPosBase_i+i < n)
	{
#ifdef DEBUG_KERNEL
		if(i==0)
		{
			printf("t%d: BLACK grid pos: (%d,%d)\n", dev, blockIdx.x, blockIdx.y);
			printf("t%d: BLACK ok3: %d, r1: %d, sign: %d\n", dev, ok3, r1, sign);
		}
#endif
  
		sgrid[ATS(i,0)] = dgrid[ATD(i,0)];
		sgrid[ATS(i,1)] = dgrid[ATD(i,1)];
		sgrid[ATS(i,2)] = dgrid[ATD(i,2)];
		if(ok3)
			sgrid[ATS(i,3)] = dgrid[ATD(i,3)];
	}  

	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: BLACK pre=============================================\n", dev);
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f- sgrid[ATS(com_i, com_j)]
				);
		else if(com_j != 2)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
	}

	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: BLACK post=============================================\n", dev);
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif		

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			dgrid[ATD(com_i, com_j)] = sgrid[ATS(com_i, com_j)];
		else if(com_j == 1)
				dgrid[ATD(com_i,1)] = sgrid[ATS(com_i,1)];
	}

}

/*
 * Red SOR through linear texturing.
 */

__global__ void rsor_lintex(float *dgrid, int n, int h, float omega, int pitch, int abs_start, size_t offset, int dev)
{
	int gridPosBase_j = blockIdx.y*2; //window size = 2
	int gridPosBase_i = blockIdx.x*(blockDim.x-2);
	int i = threadIdx.x;
	int local_w = blockDim.x;

	extern __shared__ float sgrid[];
  
  //If block can't use more than 3 rows
  int ok3 = gridPosBase_j+3 < h;
	//Red first
  int r1 = (abs_start+gridPosBase_j +1)&1;
  //Traversing sign 
  int sign = r1 + __mul24(1-r1, -1);
	//Computation indexes
	int com_i = i+1;
	int com_j = (2-r1) + __mul24(sign, 1&i);
#ifdef DEBUG_KERNEL
	if(i==0)
		printf("t%d: RED abs_pos: %d\n", dev, abs_start + gridPosBase_j +1);
#endif
  
	if(gridPosBase_i+i < n)
	{
#ifdef DEBUG_KERNEL
		if(i==0)
		{
			printf("t%d: RED grid pos: (%d,%d)\n", dev, blockIdx.x, blockIdx.y);
			printf("t%d: RED ok3: %d, r1: %d, sign: %d\n", dev, ok3, r1, sign);
		}
#endif
		
		sgrid[ATS(i,0)] = selectTex1D(dev, ATD(i,0) + offset);
		sgrid[ATS(i,1)] = selectTex1D(dev, ATD(i,1) + offset);
		sgrid[ATS(i,2)] = selectTex1D(dev, ATD(i,2) + offset);
		if(ok3)
			sgrid[ATS(i,3)] = selectTex1D(dev, ATD(i,3) + offset);
	}
  
	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("RED pre=============================================\n");
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif
		
	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
		else if(com_j != 2)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
	}

	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("RED post=============================================\n");
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			dgrid[ATD(com_i, com_j)] = sgrid[ATS(com_i, com_j)];
		else if(com_j == 1)
			dgrid[ATD(com_i, 1)] = sgrid[ATS(com_i, 1)];
	}

}

/*
 * Black SOR through linear texturing.
 */

__global__ void bsor_lintex(float *dgrid, int n, int h, float omega, int pitch, int abs_start, size_t offset, int dev)
{
	int gridPosBase_j = blockIdx.y*2; //window size = 2
	int gridPosBase_i = blockIdx.x*(blockDim.x-2);
	int i = threadIdx.x;
	int local_w = blockDim.x;

	extern __shared__ float sgrid[];
  
  int ok3 = gridPosBase_j+3 < h;
	//Red first
  int r1 = (abs_start + gridPosBase_j +1)&1;
  //Traversing sign 
  int sign = __mul24(-1,r1)+(1-r1);
	//Computation indexes
	int com_i = i+1;
	int com_j = (1+r1) + __mul24(sign,1&i);

#ifdef DEBUG_KERNEL
	if(i==0)
		printf("t%d: BLACK abs_pos: %d\n", dev, abs_start + gridPosBase_j +1);
#endif
  
	if(gridPosBase_i+i < n)
	{
	
#ifdef DEBUG_KERNEL
		if(i==0)
		{
			printf("t%d: BLACK grid pos: (%d,%d)\n", dev, blockIdx.x, blockIdx.y);
			printf("t%d: BLACK ok3: %d, r1: %d, sign: %d\n", dev, ok3, r1, sign);
		}
#endif

		sgrid[ATS(i,0)] = selectTex1D(dev, ATD(i,0));
		sgrid[ATS(i,1)] = selectTex1D(dev, ATD(i,1));
		sgrid[ATS(i,2)] = selectTex1D(dev, ATD(i,2));
		if(ok3)
			sgrid[ATS(i,3)] = selectTex1D(dev, ATD(i,3));
	}  
	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: BLACK pre=============================================\n", dev);
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f- sgrid[ATS(com_i, com_j)]
				);
		else if(com_j != 2)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
	}

	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: BLACK post=============================================\n", dev);
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif		

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			dgrid[ATD(com_i, com_j)] = sgrid[ATS(com_i, com_j)];
		else if(com_j == 1)
				dgrid[ATD(com_i,1)] = sgrid[ATS(com_i,1)];
	}

}

/*
 * Red SOR through GlobMem r/w revised for block partitioning.
 */

__global__ void rsor_gm_block(float *dgrid, int n, int h, float omega, int pitch, int abs_start_i, int abs_start_j, int dev)
{
	int gridPosBase_j = blockIdx.y*2; //window size = 2
	int gridPosBase_i = blockIdx.x*(blockDim.x-2);
	int i = threadIdx.x;
	int local_w = blockDim.x;
	
	extern __shared__ float sgrid[];
  
  //If block can't use more than 3 rows
  int ok3 = gridPosBase_j+3 < h;
	//Red first
  int r1 = ((abs_start_i+gridPosBase_i +1)&1)&((abs_start_j+gridPosBase_j +1)&1);
  r1 |= !((abs_start_i + gridPosBase_i +1)&1) & !((abs_start_j + gridPosBase_j +1)&1);
  //Traversing sign 
  int sign = r1 + __mul24(1-r1, -1);
	//Computation indexes
	int com_i = i+1;
	int com_j = (2-r1) + __mul24(sign, 1&i);
#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: RED dgrid: %d\n", dev, dgrid);
		printf("t%d: RED abs_pos_i: %d\n", dev, abs_start_i + gridPosBase_i +1);
		printf("t%d: RED abs_pos_j: %d\n", dev, abs_start_j + gridPosBase_j +1);
	}
#endif
	
	if(gridPosBase_i+i < n)
	{
	
#ifdef DEBUG_KERNEL
		if(i==0)
		{
			printf("t%d: RED grid pos: (%d,%d)\n", dev, blockIdx.x, blockIdx.y);
			printf("t%d: RED ok3: %d, r1: %d, sign: %d\n", dev, ok3, r1, sign);
		}
#endif
		
		sgrid[ATS(i,0)] = dgrid[ATD(i,0)];
		sgrid[ATS(i,1)] = dgrid[ATD(i,1)];
		sgrid[ATS(i,2)] = dgrid[ATD(i,2)];
		if(ok3)
			sgrid[ATS(i,3)] = dgrid[ATD(i,3)];
	}
  
	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("RED pre=============================================\n");
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif
		
	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
		else if(com_j != 2)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
	}

	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("RED post=============================================\n");
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			dgrid[ATD(com_i, com_j)] = sgrid[ATS(com_i, com_j)];
		else if(com_j == 1)
			dgrid[ATD(com_i, 1)] = sgrid[ATS(com_i, 1)];
	}

}

/*
 * Black SOR through GlobMem r/w revised for block partitioning.
 */

__global__ void bsor_gm_block(float *dgrid, int n, int h, float omega, int pitch, int abs_start_i, int abs_start_j, int dev)
{
	int gridPosBase_j = blockIdx.y*2; //window size = 2
	int gridPosBase_i = blockIdx.x*(blockDim.x-2);
	int i = threadIdx.x;
	int local_w = blockDim.x;
	extern __shared__ float sgrid[];
  
  int ok3 = gridPosBase_j+3 < h;
	//Red first
  int r1 = ((abs_start_i + gridPosBase_i +1)&1) & ((abs_start_j + gridPosBase_j +1)&1);
  r1 |= !((abs_start_i + gridPosBase_i +1)&1) & !((abs_start_j + gridPosBase_j +1)&1);
  //Traversing sign 
  int sign = __mul24(-1,r1)+(1-r1);
	//Computation indexes
	int com_i = i+1;
	int com_j = (1+r1) + __mul24(sign,1&i);

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: BLACK abs_pos_i: %d\n", dev, abs_start_i + gridPosBase_i +1);
		printf("t%d: BLACK abs_pos_j: %d\n", dev, abs_start_j + gridPosBase_j +1);
	}
#endif
  
	if(gridPosBase_i+i < n)
	{
#ifdef DEBUG_KERNEL
		if(i==0)
		{
			printf("t%d: BLACK grid pos: (%d,%d)\n", dev, blockIdx.x, blockIdx.y);
			printf("t%d: BLACK ok3: %d, r1: %d, sign: %d\n", dev, ok3, r1, sign);
		}
#endif
  
		sgrid[ATS(i,0)] = dgrid[ATD(i,0)];
		sgrid[ATS(i,1)] = dgrid[ATD(i,1)];
		sgrid[ATS(i,2)] = dgrid[ATD(i,2)];
		if(ok3)
			sgrid[ATS(i,3)] = dgrid[ATD(i,3)];
	}  

	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: BLACK pre=============================================\n", dev);
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f- sgrid[ATS(com_i, com_j)]
				);
		else if(com_j != 2)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
	}

	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: BLACK post=============================================\n", dev);
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif		

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			dgrid[ATD(com_i, com_j)] = sgrid[ATS(com_i, com_j)];
		else if(com_j == 1)
				dgrid[ATD(com_i,1)] = sgrid[ATS(com_i,1)];
	}

}

/*
 * Red SOR through linear texturing revised for block partitioning.
 */

__global__ void rsor_lintex_block(float *dgrid, int n, int h, float omega, int pitch, int abs_start_i, int abs_start_j, size_t offset, int dev)
{
	int gridPosBase_j = blockIdx.y*2; //window size = 2
	int gridPosBase_i = blockIdx.x*(blockDim.x-2);
	int i = threadIdx.x;
	int local_w = blockDim.x;

	extern __shared__ float sgrid[];
  
  //If block can't use more than 3 rows
  int ok3 = gridPosBase_j+3 < h;
	//Red first
  int r1 = ((abs_start_i + gridPosBase_i +1)&1)&((abs_start_j + gridPosBase_j +1)&1);
  r1 |= !((abs_start_i + gridPosBase_i +1)&1) & !((abs_start_j + gridPosBase_j +1)&1);
  //Traversing sign 
  int sign = r1 + __mul24(1-r1, -1);
	//Computation indexes
	int com_i = i+1;
	int com_j = (2-r1) + __mul24(sign, 1&i);
#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: RED dgrid: %d\n", dev, dgrid);
		printf("t%d: RED abs_pos_i: %d\n", dev, abs_start_i + gridPosBase_i +1);
		printf("t%d: RED abs_pos_j: %d\n", dev, abs_start_j + gridPosBase_j +1);
	}
#endif
  
	if(gridPosBase_i+i < n)
	{
#ifdef DEBUG_KERNEL
		if(i==0)
		{
			printf("t%d: RED grid pos: (%d,%d)\n", dev, blockIdx.x, blockIdx.y);
			printf("t%d: RED ok3: %d, r1: %d, sign: %d\n", dev, ok3, r1, sign);
		}
#endif
		
		sgrid[ATS(i,0)] = selectTex1D(dev, ATD(i,0)+offset);
		sgrid[ATS(i,1)] = selectTex1D(dev, ATD(i,1)+offset);
		sgrid[ATS(i,2)] = selectTex1D(dev, ATD(i,2)+offset);
		if(ok3)
			sgrid[ATS(i,3)] = selectTex1D(dev, ATD(i,3)+offset);
	}
  
	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("RED pre=============================================\n");
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif
		
	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
		else if(com_j != 2)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
	}

	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("RED post=============================================\n");
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			dgrid[ATD(com_i, com_j)] = sgrid[ATS(com_i, com_j)];
		else if(com_j == 1)
			dgrid[ATD(com_i, 1)] = sgrid[ATS(com_i, 1)];
	}

}

/*
 * Black SOR through linear texturing revised for block partitioning.
 */

__global__ void bsor_lintex_block(float *dgrid, int n, int h, float omega, int pitch, int abs_start_i, int abs_start_j, size_t offset, int dev)
{
	int gridPosBase_j = blockIdx.y*2; //window size = 2
	int gridPosBase_i = blockIdx.x*(blockDim.x-2);
	int i = threadIdx.x;
	int local_w = blockDim.x;

	extern __shared__ float sgrid[];
  
  int ok3 = gridPosBase_j+3 < h;
	//Red first
  int r1 = ((abs_start_i + gridPosBase_i +1)&1)&((abs_start_j + gridPosBase_j +1)&1);
  r1 |= !((abs_start_i + gridPosBase_i +1)&1) & !((abs_start_j + gridPosBase_j +1)&1);
  //Traversing sign 
  int sign = __mul24(-1,r1)+(1-r1);
	//Computation indexes
	int com_i = i+1;
	int com_j = (1+r1) + __mul24(sign,1&i);

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: BLACK abs_pos_i: %d\n", dev, abs_start_i + gridPosBase_i +1);
		printf("t%d: BLACK abs_pos_j: %d\n", dev, abs_start_j + gridPosBase_j +1);
	}
#endif
  
	if(gridPosBase_i+i < n)
	{
	
#ifdef DEBUG_KERNEL
		if(i==0)
		{
			printf("t%d: BLACK grid pos: (%d,%d)\n", dev, blockIdx.x, blockIdx.y);
			printf("t%d: BLACK ok3: %d, r1: %d, sign: %d\n", dev, ok3, r1, sign);
		}
#endif

		sgrid[ATS(i,0)] = selectTex1D(dev, ATD(i,0)+offset);
		sgrid[ATS(i,1)] = selectTex1D(dev, ATD(i,1)+offset);
		sgrid[ATS(i,2)] = selectTex1D(dev, ATD(i,2)+offset);
		if(ok3)
			sgrid[ATS(i,3)] = selectTex1D(dev, ATD(i,3)+offset);
	}  
	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: BLACK pre=============================================\n", dev);
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f- sgrid[ATS(com_i, com_j)]
				);
		else if(com_j != 2)
			sgrid[ATS(com_i, com_j)] += 
				omega*(
				(sgrid[ATS(com_i-1, com_j)] + sgrid[ATS(com_i+1, com_j)] + sgrid[ATS(com_i, com_j-1)] + sgrid[ATS(com_i, com_j+1)])/4.f - sgrid[ATS(com_i, com_j)]
				);
	}

	__syncthreads();

#ifdef DEBUG_KERNEL
	if(i==0)
	{
		printf("t%d: BLACK post=============================================\n", dev);
		int limit_y = ok3 ? 4 : 3;
		for(int s=0; s<limit_y; s++)
		{
			int limit = n > local_w ? local_w : n;
			for(int t=0; t<limit; t++)
			{
				printf("%f ", sgrid[ATS(t,s)]);
			}
			printf("\n");
		}
	}
#endif		

	if((i<local_w-2)&&(gridPosBase_i+i<n-2))
	{
		if(ok3)
			dgrid[ATD(com_i, com_j)] = sgrid[ATS(com_i, com_j)];
		else if(com_j == 1)
				dgrid[ATD(com_i,1)] = sgrid[ATS(com_i,1)];
	}

}
#endif
