#ifndef __PSYNC_CU
#define __PSYNC_CU

#include <pthread.h>
#include <math.h>

#include <sys/types.h>
#include <time.h>

#include <pde_kernels.cu>

struct Params
{
	char solver;
	float omega;
	char threads;
	int device;
	float *grid;
	int n;
	int h;
	int stride;
	int border_size;
	int iterations;
};

pthread_mutex_t *mutex;

pthread_cond_t *rCond, *wCond;
int *rAcc, *wAcc;

//Timing arrays. The prefix ext_ means outside the core computation.
double *ext_cpy_time;
double *ext_sync_time;
double *cpy_time;
double *sync_time;
double *comm_time;
double *dev_sel_time; //Device selection time

////////////////////////// CUDA WRAPPERS //////////////////////////////////////
extern "C"
	int call_solver(int solver, float *omega, float *grid, int m, int n, int iterations, int threads, int border_size);
///////////////////////////////////////////////////////////////////////////////

/*
 * checkCUDAError
 *	msg - String message to show in case an error occurs.
 */

void checkCUDAError(const char *msg)
{
	cudaError_t err = cudaGetLastError();
	if(cudaSuccess != err)
	{
		fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString(err));
	}
}

/*
 * getTime
 * Based on the timespec format. Returns end-start in seconds.
 *	
 */

double getTime(timespec start, timespec end)
{
	long nsec = end.tv_nsec-start.tv_nsec;
	if(nsec < 0)
	{
		nsec = 1E9+nsec;
		end.tv_sec-=1;
	}
	return (double)(end.tv_sec-start.tv_sec)+(double)(nsec*1E-9);
}

/*
 * printPortion
 *	
 */

void printPortion(int who, int step, float *m, int w, int h, const char *msg)
{
	printf("Worker %d's portion at step %d =========.\n", who, step);
	printf("Msg: %s.\n", msg);
	for(int j = 0; j < h; j++)
	{
		for(int i = 0; i < w; i++)
			printf("%.2f ", m[j*w+i]);
		printf("\n");
	}
	printf("=====================================\n");
}

/*
 * selectTextureToBind
 * Select the right texture reference depending on the GPU context.
 *	
 */

size_t selectTextureToBind(int device, int step, float *devPtr, size_t size)
{
	size_t offset;
	switch(device)
	{
		case 0:
			cudaBindTexture(&offset, texRef_0, devPtr, size);
			break;
		case 1:
			cudaBindTexture(&offset, texRef_1, devPtr, size);
			break;
		case 2:
			cudaBindTexture(&offset, texRef_2, devPtr, size);
			break;
		case 3:
			cudaBindTexture(&offset, texRef_3, devPtr, size);
			break;
		default:
			printf("Eventually you must add textures to bind to...\n");
	}
	
	return offset;
}

/*
 * selectDevice
 *	selects the specified device.
 */

void selectDevice(int device)
{

	cudaSetDevice(device);

	int dev;
	cudaDeviceProp deviceProp;

	cudaGetDevice(&dev);
	cudaGetDeviceProperties(&deviceProp, dev);

	printf("t%d Running on device %d: %s.\n", device, dev, deviceProp.name);
}

/*
 * getMaxDeviceSelectionTime
 */
 
double getMaxDeviceSelectionTime(int threads)
{
	double max = dev_sel_time[0];
	for(int i = 1; i < threads; i++)
		if(dev_sel_time[i] > max)
			max = dev_sel_time[i];
	
	return max;
}
#endif
