#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>

#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>

#include <sys/time.h>
#include <sys/resource.h>

#include <errno.h>

#define MAX_TEMP 100
#define K_1GiB 1024*1024*1024.l

//command-line options (short version)
const char * const short_options = "hm:n:f:";

//command-line options (long version)
const struct option long_options[] = {
	{"help",				0,	NULL,	'h'},
	{"rows",				1,	NULL,	'm'},
	{"columns",			1,	NULL,	'n'},
	{"file",				1,	NULL,	'f'}
};

/*
 * Print usage information to <stream> and exit with <exit_code>.
 */
void print_usage(FILE* stream, int exit_code)
{
	fprintf(stream, "Usage: grid_gen -m <rows> -n <columns> -f <file> [-h]\n");
	
	exit(exit_code);
}

/*
 * MAIN FUNCTION
 */
int main(int argc, char **argv)
{
	size_t m, n;
	const char* outfile = NULL;
	int next_opt, opt_counter = 0;
	
	do{
		next_opt = getopt_long(argc, argv, short_options, long_options, NULL);
		
		switch(next_opt)
		{
			case 'h':
				print_usage(stdout, 0);
			case 'm':
				m = atoi(optarg);
				opt_counter++;
				break;
			case 'n':
				n = atoi(optarg);
				opt_counter++;
				break;
			case 'f':
				outfile = optarg;
				opt_counter++;
				break;
			case '?':
				fprintf(stderr, "Unexpected option.\n");
				print_usage(stderr, 1);
			case -1:
				//Done with options
				break;
			default:
				exit(-1);
		}
	} while(next_opt != -1);
	
	if(opt_counter != 3)
	{
		fprintf(stderr, "Missing mandatory options.\n");
		print_usage(stderr, 1);
	}

	size_t size = m*n;
	
	size_t iter = (size*sizeof(float))/(K_1GiB);
	
	size_t rest = size*sizeof(float)-iter*(K_1GiB);

	printf("Requested size:%ld.\nNumber of 1GB iterations: %ld.\nRest: %ld.\n", size*sizeof(float), iter, rest);
	
	int fd;

	if((fd = open(outfile, O_CREAT | O_RDWR | O_TRUNC,  S_IRUSR | S_IWUSR | S_IRGRP)) == -1)
	{
		fprintf(stderr, "Error opening file.\n");
		return -1;
	}

	float *grid;

	grid = (float *)calloc(size, sizeof(float));
	printf("Mem callocated.\n");

	for(size_t j=0; j<m; j++)
	{
		grid[j*n] = MAX_TEMP;
		grid[j*n+n-1] = MAX_TEMP;
	}
	for(size_t i=0; i<n; i++)
	{
		grid[i] = MAX_TEMP;
		grid[(m-1)*n+i] = MAX_TEMP;
	}
	
	ssize_t r;
	size_t total = 0;
	for(size_t i = 0; i < iter; i++)
	{
		if((r=pwrite(fd, grid, K_1GiB, i*K_1GiB)) != K_1GiB)
		{
			fprintf(stderr, "At iteration %ld pwrite returns %ld.\n", i, r);
			perror(NULL);
			return -1;
		}
		total += r;
	}
	
	if((r=pwrite(fd, grid, rest, total)) != rest)
	{
		fprintf(stderr, "Writing rest pwrite returns %ld.\n", r);
		perror(NULL);
		return -1;
	}
	total += r;

	struct stat fs;

	fstat(fd, &fs);
	
	if(fs.st_size != size*sizeof(float))
	{
		printf("Error: size differs from what expected. Expected: %ld, Size: %ld, # allocated blocks: %ld\n", size*sizeof(float), fs.st_size, fs.st_blocks);
		exit(EXIT_FAILURE);
	}

	printf("done. Size: %ld, # allocated blocks: %ld\n", fs.st_size, fs.st_blocks);
	printf("float size: %ld.\n", sizeof(float));
	
	close(fd);
	free(grid);

	exit(EXIT_SUCCESS);
}
