/**
 * The module provides wrapper functions to the pde kernels when the domain is strip partitioned.
 *											 					 =====												 
 *	Assume a cartesian communicator:					 |0|1|
 *																 =====
 *											 					 |2|3|
 *											 					 =====		
 *										 
 **/

#include <psync.cu>

/*
 * moveBlockToDevice
 * move a block of data to the GPU
 *	
 */

void moveBlockToDevice(int device, int neighbor, int step, int direction, float *hostPtr, size_t hPitch, float *devPtr, size_t dPitch, size_t bytes, size_t rows)
{
	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	
	float temp_time;
	timespec h_start, h_end;
	clock_gettime(CLOCK_REALTIME, &h_start);
	
	pthread_mutex_lock(&mutex[neighbor+direction]);
		while(wAcc[neighbor+direction] != step+1)
		{
			pthread_cond_wait(&wCond[neighbor+direction], &mutex[neighbor+direction]);
		}
	pthread_mutex_unlock(&mutex[neighbor+direction]);

	clock_gettime(CLOCK_REALTIME, &h_end);
	sync_time[device] += getTime(h_start, h_end);

	cudaEventRecord(start, 0);	

	cudaMemcpy2D(devPtr, dPitch, hostPtr, hPitch, bytes, rows, cudaMemcpyHostToDevice);

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop); 
	
	cudaEventElapsedTime(&temp_time, start, stop);
	cpy_time[device] += temp_time;

	clock_gettime(CLOCK_REALTIME, &h_start);

	pthread_mutex_lock(&mutex[neighbor+direction]);
		rAcc[neighbor+direction]++;
	pthread_mutex_unlock(&mutex[neighbor+direction]);
	pthread_cond_signal(&rCond[neighbor+direction]);

	clock_gettime(CLOCK_REALTIME, &h_end);
	sync_time[device] += getTime(h_start, h_end);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);
}

/*
 * moveBlockToHost
 * move a block of data to the host
 *	
 */

void moveBlockToHost(int device, int neighbor, int step, int direction, float *hostPtr, size_t hPitch, float *devPtr, size_t dPitch, size_t bytes, size_t rows)
{
	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	
	float temp_time;
	timespec h_start, h_end;
	clock_gettime(CLOCK_REALTIME, &h_start);

	pthread_mutex_lock(&mutex[device+direction]);
		while(rAcc[device+direction] != step+1)
		{
			pthread_cond_wait(&rCond[device+direction], &mutex[device+direction]);
		}
	pthread_mutex_unlock(&mutex[device+direction]);

	clock_gettime(CLOCK_REALTIME, &h_end);
	sync_time[device] += getTime(h_start, h_end);

	cudaEventRecord(start, 0);	

	cudaMemcpy2D(hostPtr, hPitch, devPtr, dPitch, bytes, rows, cudaMemcpyDeviceToHost);

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop); 
	cudaEventElapsedTime(&temp_time, start, stop);
	cpy_time[device] += temp_time;

	clock_gettime(CLOCK_REALTIME, &h_start);

	pthread_mutex_lock(&mutex[device+direction]);
		wAcc[device+direction]++;
	pthread_mutex_unlock(&mutex[device+direction]);
	pthread_cond_signal(&wCond[device+direction]);	

	clock_gettime(CLOCK_REALTIME, &h_end);
	sync_time[device] += getTime(h_start, h_end);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);
}

/*
 * tetra_thread
 *	params - contains the parameters required to setup a partial solver (i.e. the solver running on one of the GPU and solving 
 *	computing a specific portion of domain).
 */

void * tetra_thread(void *params)
{
	//Cast the cookie pointer to the right type
	struct Params *p = (struct Params *)params;
	
	//Vertical and horizontal neighbors
	int vn, hn;
	
	if(p->device%2)
		hn = p->device-1;
	else
		hn = p->device+1;

	if(p->device <= 1)
		vn = p->device+2;
	else
		vn = p->device-2;

	struct timespec h_start, h_end;

	clock_gettime(CLOCK_REALTIME, &h_start);
	
	selectDevice(p->device);

	clock_gettime(CLOCK_REALTIME, &h_end);

	printf("device selection time : %.9fs\n", dev_sel_time[p->device] = getTime(h_start, h_end));

	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	
	float r_kernel_time = 0, b_kernel_time = 0, temp_time;
	
	//Setup subdomain processing
	float *dgrid;
	int offset, height;
	size_t pitch;

	//Total number of rows counting borders. GPUs in the middle
	//deal with two sides.
	height = p->h+(p->border_size);
	
	//Offset to the beginning of the subdomain
	switch(p->device)
	{
		case 0:
			offset = 0;
			break;
		case 1:
			offset = p->border_size;
			break;
		case 2:
			offset = p->border_size*p->stride;
			break;
		case 3:
			offset = p->border_size*p->stride+p->border_size;
	}
		
	//Allocate memory on the device
	cudaMallocPitch((void **)&dgrid, &pitch, (p->n+p->border_size)*sizeof(float), height);
	printf("t%d pitch: %ld, height: %d, offset: %d.\n", p->device, pitch, height, offset);
	printf("t%d grid: %p, dgrid: %p.\n", p->device, p->grid, dgrid);
	
	//Kernel Execution Configuration
	int grid_x, grid_y, local_w;

	//Number of threads per threadblock
	local_w = 112;
	printf("t%d local_w: %d, stride: %d\n", p->device, local_w, p->stride);

	dim3 blockDim(local_w);

	//First transfer. Entire subdomain+borders.
	comm_time[p->device] = ext_cpy_time[p->device] = cpy_time[p->device] = 0;

	cudaEventRecord(start, 0);	

	cudaMemcpy2D(dgrid, pitch, p->grid-offset, p->stride*sizeof(float), (p->n+p->border_size)*sizeof(float), height, cudaMemcpyHostToDevice);

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&temp_time, start, stop);
	ext_cpy_time[p->device] += temp_time;
	printf("t%d copied first block in %f ms.\n", p->device, ext_cpy_time[p->device]);

	clock_gettime(CLOCK_REALTIME, &h_start);

	//Increase read counters.
	pthread_mutex_lock(&mutex[vn]);
		rAcc[vn]++;
	pthread_mutex_unlock(&mutex[vn]);
	pthread_cond_signal(&rCond[vn]);
	
	//Horizontal mutexes are in position i > 3
	pthread_mutex_lock(&mutex[hn+4]);
		rAcc[hn+4]++;
	pthread_mutex_unlock(&mutex[hn+4]);
	pthread_cond_signal(&rCond[hn+4]);

	clock_gettime(CLOCK_REALTIME, &h_end);
	ext_sync_time[p->device] = getTime(h_start, h_end);
	sync_time[p->device] = 0;
	comm_time[p->device] += sync_time[p->device];

// CORE COMPUTATION - Begin //////////////////////////

	//So to be able to iterate with border_size == 0
	int b = p->border_size > 0 ? p->border_size : 1; 
	
	for(int i = 0; i < p->iterations/b; i++)
	{	
		int border = b;
		//k_* are offsets used to point to the beginning of the area to process
		int k_v=0, k_h=0;
		while(border > 0)
		{
			//Compute grid dimensions.
			//Horizontally, blocks shift with window size local_w-2.
			//Vertically, blocks shift with window size 2.
			grid_x = (p->n+border)/(local_w-2);
			if(grid_x*(local_w-2) < p->n+border) grid_x++;

			grid_y = (2*height-4)/sizeof(float) + (height)%2;

			dim3 gridDim(grid_x, grid_y);
			
			switch(p->solver)
			{
				case 0:
				{	
					/* apply R/B SOR, using global mem for fetching data. */					
					cudaEventRecord(start, 0);	

					rsor_gm_block<<<gridDim, blockDim, 4*local_w*sizeof(float)>>>(dgrid+k_v*(pitch/sizeof(float))+k_h, p->n+border, height, p->omega, 
						pitch/sizeof(float), (p->device%2)*((p->stride-p->n)-border), (p->device>1)*(p->h-border), p->device);

					cudaEventRecord(stop, 0);
					cudaEventSynchronize(stop);
					cudaEventElapsedTime(&temp_time, start, stop);
					r_kernel_time += temp_time;

					cudaEventRecord(start, 0);	

					bsor_gm_block<<<gridDim, blockDim, 4*local_w*sizeof(float)>>>(dgrid+k_v*(pitch/sizeof(float))+k_h, p->n+border, height, p->omega, 
						pitch/sizeof(float), (p->device%2)*((p->stride-p->n)-border), (p->device>1)*(p->h-border), p->device); 

					cudaEventRecord(stop, 0);
					cudaEventSynchronize(stop);

					cudaEventElapsedTime(&temp_time, start, stop);
					b_kernel_time += temp_time;

					break;
				}
				case 1:
				{	
					/* apply R/B SOR, using a texture to fetch data */
					size_t tex_offset = selectTextureToBind(p->device, i, dgrid+k_v*(pitch/sizeof(float))+k_h, height*pitch);

					cudaEventRecord(start, 0);	

					rsor_lintex_block<<<gridDim, blockDim, 4*local_w*sizeof(float)>>>(dgrid+k_v*(pitch/sizeof(float))+k_h, p->n+border, height, p->omega, 
					pitch/sizeof(float), (p->device%2)*(p->n-border), (p->device>1)*(p->h-border), tex_offset/sizeof(float), p->device); 

					cudaEventRecord(stop, 0);
					cudaEventSynchronize(stop);

					cudaEventElapsedTime(&temp_time, start, stop);
					r_kernel_time += temp_time;

					cudaEventRecord(start, 0);	

					bsor_lintex_block<<<gridDim, blockDim, 4*local_w*sizeof(float)>>>(dgrid+k_v*(pitch/sizeof(float))+k_h, p->n+border, height, p->omega, 
						pitch/sizeof(float), (p->device%2)*(p->n-border), (p->device>1)*(p->h-border), tex_offset/sizeof(float), p->device); 

					cudaEventRecord(stop, 0);
					cudaEventSynchronize(stop);
					cudaEventElapsedTime(&temp_time, start, stop);
					b_kernel_time += temp_time;

					break;
				}
			}

			//Eventually increase offsets
			border--; height--; k_v += 1*(p->device > 1);	k_h += (p->device%2);
		} //End border consuming while
		
		//Restore height
		height += b;

//***** Communication *****
#ifdef TIMER
		clock_gettime(CLOCK_REALTIME, &h_start);
#endif			
		switch(p->device)
		{
			case 0:
				moveBlockToHost(0, 2, i, 0/*vertical*/, p->grid+(p->h-p->border_size)*p->stride, p->stride*sizeof(float), 
					dgrid+(p->h-p->border_size)*pitch/sizeof(float), pitch, p->n*sizeof(float), p->border_size);
				moveBlockToDevice(0, 2, i, 0/*vertical*/, p->grid+(p->h)*p->stride, p->stride*sizeof(float), dgrid+(p->h)*pitch/sizeof(float),
					pitch, p->n*sizeof(float), p->border_size);
				moveBlockToHost(0, 1, i, 4/*horizontal*/, p->grid+(p->n-p->border_size), p->stride*sizeof(float), dgrid+(p->n-p->border_size),
					pitch, p->border_size*sizeof(float), p->h-p->border_size);
				moveBlockToDevice(0, 1, i, 4/*horizontal*/, p->grid+p->n, p->stride*sizeof(float), dgrid+(p->n),
					pitch, p->border_size*sizeof(float), p->h+p->border_size);
				break;
			case 1:
				moveBlockToHost(1, 3, i, 0/*vertical*/, p->grid+(p->h-p->border_size)*p->stride, p->stride*sizeof(float), 
					dgrid+(p->h-p->border_size)*pitch/sizeof(float)+p->border_size, pitch, p->n*sizeof(float), p->border_size);
				moveBlockToDevice(1, 3, i, 0/*vertical*/, p->grid+(p->h)*p->stride, p->stride*sizeof(float), dgrid+(p->h)*pitch/sizeof(float)+p->border_size,
					pitch, p->n*sizeof(float), p->border_size);
				moveBlockToHost(1, 0, i, 4/*horizontal*/, p->grid, p->stride*sizeof(float), dgrid+p->border_size,
					pitch, p->border_size*sizeof(float), p->h-p->border_size);
				moveBlockToDevice(1, 0, i, 4/*horizontal*/, p->grid-p->border_size, p->stride*sizeof(float), dgrid,
					pitch, p->border_size*sizeof(float), p->h+p->border_size);
				break;
			case 2:
				moveBlockToHost(2, 0, i, 0/*vertical*/, p->grid, p->stride*sizeof(float), dgrid+p->border_size*pitch/sizeof(float),
					pitch, p->n*sizeof(float), p->border_size);
				moveBlockToDevice(2, 0, i, 0/*vertical*/, p->grid-p->border_size*p->stride, p->stride*sizeof(float), dgrid,
					pitch, p->n*sizeof(float), p->border_size);
				moveBlockToHost(2, 3, i, 4/*horizontal*/, p->grid+p->border_size*p->stride+p->n-p->border_size, p->stride*sizeof(float), 
				dgrid+(2*p->border_size)*pitch/sizeof(float)+p->n-p->border_size, pitch, p->border_size*sizeof(float), p->h-p->border_size);
				moveBlockToDevice(2, 3, i, 4/*horizontal*/, p->grid-(p->border_size*p->stride)+p->n, p->stride*sizeof(float), dgrid+(p->n),
					pitch, p->border_size*sizeof(float), p->h+p->border_size);
				break;
			case 3:
				moveBlockToHost(3, 1, i, 0/*vertical*/, p->grid, p->stride*sizeof(float), dgrid+p->border_size*pitch/sizeof(float)+p->border_size,
					pitch, p->n*sizeof(float), p->border_size);
				moveBlockToDevice(3, 1, i, 0/*vertical*/, p->grid-p->border_size*p->stride, p->stride*sizeof(float), dgrid+p->border_size,
					pitch, p->n*sizeof(float), p->border_size);
				moveBlockToHost(3, 2, i, 4/*horizontal*/, p->grid+p->border_size*p->stride, p->stride*sizeof(float), 
				dgrid+(2*p->border_size)*pitch/sizeof(float)+p->border_size, pitch, p->border_size*sizeof(float), p->h-p->border_size);
				moveBlockToDevice(3, 2, i, 4/*horizontal*/, p->grid-(p->border_size*p->stride)-p->border_size, p->stride*sizeof(float), dgrid,
					pitch, p->border_size*sizeof(float), p->h+p->border_size);
				break;
		}

		clock_gettime(CLOCK_REALTIME, &h_end);
		comm_time[p->device] += getTime(h_start, h_end);

//***** End Communication *****
	}

// CORE COMPUTATION - End //////////////////////////
	
	//Last transfer. Just subdomain
	switch(p->device)
	{
		case 0:
			offset = 0;
			break;		
		case 1:
			offset = p->border_size;
			break;		
		case 2:
			offset = p->border_size*pitch/sizeof(float);
			break;		
		case 3:
			offset = p->border_size*pitch/sizeof(float)+p->border_size;
			break;		
	}

		cudaEventRecord(start, 0);	

		cudaMemcpy2D(p->grid, p->stride*sizeof(float), dgrid+offset, pitch, p->n*sizeof(float), p->h, cudaMemcpyDeviceToHost);

		cudaEventRecord(stop, 0);
		cudaEventSynchronize(stop);
		cudaEventElapsedTime(&temp_time, start, stop);
		ext_cpy_time[p->device] += temp_time;

	//Format and display output
	printf("t%d avarage RSOR kernel time: %f ms.\n", p->device, r_kernel_time/p->iterations);
	printf("t%d avarage BSOR kernel time: %f ms.\n", p->device, b_kernel_time/p->iterations);
	printf("t%d total kernel time: %f ms.\n", p->device, r_kernel_time+b_kernel_time);
	printf("t%d MKT: %f ms.\n", p->device, (r_kernel_time+b_kernel_time)/p->iterations);
	
	printf("t%d total transfer time: %f ms [iter: %f + ext: %f].\n", p->device, cpy_time[p->device]+ext_cpy_time[p->device], 
		cpy_time[p->device], ext_cpy_time[p->device]);
	printf("t%d MTT: %f ms.\n", p->device, cpy_time[p->device]/(p->iterations/b));
	printf("t%d total synchronization time: %f ms [iter: %f + ext: %f].\n", p->device, (sync_time[p->device]+ext_sync_time[p->device])*1E3, 
		sync_time[p->device]*1E3, ext_sync_time[p->device]*1E3);
	printf("t%d MST: %f ms.\n", p->device, (sync_time[p->device]*1E3)/(p->iterations/b));
	printf("t%d total communication time: %f ms [iter: %f + ext: %f].\n", p->device, 
		comm_time[p->device]*1E3+ext_sync_time[p->device]*1E3+ext_cpy_time[p->device], comm_time[p->device]*1E3, 
		ext_sync_time[p->device]*1E3+ext_cpy_time[p->device]);
	printf("t%d MCT: %f ms.\n", p->device, (comm_time[p->device]*1E3)/(p->iterations/b));

	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	cudaFree(dgrid);	
	
	return NULL;
}

/*
 * call_solver
 * 	solver - the index of the solver to use.
 * 	grid - pointer to the PDE domain.
 *		n - width/height of the domain in terms of number of elements per row/column.
 *		iterations - the number of times the solver should iterate.
 *		threads - indicates how many threads, and finally devices, should cooperate in solving the PDE.
 *	
 * 	Instantiates the right number of threads and manages their cooperation to solve the PDE.
 */

int call_solver(int solver, float *omega, float *grid, int m, int n, int iterations, int threads, int border_size)
{
	pthread_t *t_id = new pthread_t[threads];
	
	struct Params *t_p = new struct Params[threads];

	struct timespec start, end;
	
	mutex = new pthread_mutex_t[2*(threads)];
	rCond = new pthread_cond_t[2*(threads)];
	wCond = new pthread_cond_t[2*(threads)];
	rAcc = new int[2*(threads)];
	wAcc = new int[2*(threads)];
	
	ext_cpy_time = new double[threads];
	ext_sync_time = new double[threads];
	cpy_time = new double[threads];
	sync_time = new double[threads];
	comm_time = new double[threads];
	dev_sel_time = new double[threads];

	printf("border_size: %d.\n", border_size);

	if(*omega == 0)
	{
		float pi = 4.0*atan(1.0);
		printf("pi: %.15f\n", pi);
		float interval = 1.0/(n+1);
		*omega = 2.0/(1.0+sin(pi*interval));
		if(*omega >1.9) *omega = 1.9;
	}

	printf("omega: %f.\n", *omega);
	
	for(int i=0; i<2*(threads); i++)
	{
		pthread_mutex_init(&mutex[i], NULL); 
		pthread_cond_init(&rCond[i], NULL); 
		pthread_cond_init(&wCond[i], NULL); 
		rAcc[i] = 0;
		wAcc[i] = 0;
	}

	for(int i=0; i<threads; i++)
	{
		t_p[i].solver = solver;
		t_p[i].omega = *omega;
		t_p[i].stride = n;
		t_p[i].n = n/2;
		t_p[i].h = m/2;
		t_p[i].border_size = border_size;
		t_p[i].iterations = iterations;
		t_p[i].grid = grid + t_p[i].h*n*(i>1) + t_p[i].n*(i%2 != 0);
		t_p[i].device = i;
		t_p[i].threads = threads;
	}
	
	if(t_p[0].h*threads < m)
		t_p[2].h = (t_p[3].h += m-(t_p[0].h*threads));
	
	if(t_p[0].n*threads < n)
		t_p[0].n = (t_p[1].n += n-(t_p[0].n*threads));

	clock_gettime(CLOCK_REALTIME, &start);
	
	for(int i=0; i<threads; i++)
		pthread_create(&t_id[i], NULL, &tetra_thread, &t_p[i]);

	for(int i=0; i<threads; i++)
	{
		pthread_join(t_id[i], NULL);
		printf("t%d joined.\n", i);		
	}

	clock_gettime(CLOCK_REALTIME, &end);

	double max_dev_sel = getMaxDeviceSelectionTime(threads);
	printf("Elapsed time: %.9fs\n", getTime(start, end)-max_dev_sel);

	for(int i=0; i<2*threads; i++)
	{
		pthread_mutex_destroy(&mutex[i]);
		pthread_cond_destroy(&rCond[i]);
		pthread_cond_destroy(&wCond[i]);
	}

	delete [] mutex;
	delete [] rCond;
	delete [] wCond;
	delete [] rAcc;
	delete [] wAcc;

	delete [] t_id;
	delete [] t_p;

	delete [] ext_cpy_time;
	delete [] ext_sync_time;
	delete [] cpy_time;
	delete [] sync_time;
	delete [] comm_time;
	delete [] dev_sel_time;
	
	return 0;
}
