/**
 * The module provides wrapper functions to the pde kernels when the domain is strip partitioned.
 **/

#include <psync.cu>

/*
 * write_ghosts
 *	Copies a shared area from device to host.
 */

void write_ghosts(int device, int block, int neighbor, int step, float *host_addr, size_t host_pitch, float *dev_addr, size_t dev_pitch, int width_byte, int height)
{

	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	
	float temp_time;
	timespec h_start, h_end;
	clock_gettime(CLOCK_REALTIME, &h_start);

	//Lock mutex to check condition variable. Proceed just if the neighbor 
	//read the border during the previous iteration
	pthread_mutex_lock(&mutex[block]);
		while(rAcc[block] != step+1)
		{
			pthread_cond_wait(&rCond[block], &mutex[block]);
		}
	pthread_mutex_unlock(&mutex[block]);

	clock_gettime(CLOCK_REALTIME, &h_end);
	sync_time[device] += getTime(h_start, h_end);
	cudaEventRecord(start, 0);	

	cudaMemcpy2D(host_addr, host_pitch, dev_addr, dev_pitch, width_byte, height, cudaMemcpyDeviceToHost);

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop); //Block until the event is actually recorded
	
	cudaEventElapsedTime(&temp_time, start, stop);
	cpy_time[device] += temp_time;

	clock_gettime(CLOCK_REALTIME, &h_start);

	//Lock mutex to increase the write counter.
	pthread_mutex_lock(&mutex[block]);
		wAcc[block]++;
	pthread_mutex_unlock(&mutex[block]);
	pthread_cond_signal(&wCond[block]);

	clock_gettime(CLOCK_REALTIME, &h_end);
	sync_time[device] += getTime(h_start, h_end);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);
}

/*
 * read_ghosts
 *	Copies a shared area from host to device.
 */

void read_ghosts(int device, int block, int neighbor, int step, float *host_addr, size_t host_pitch, float *dev_addr, size_t dev_pitch, int width_byte, int height)
{
	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	
	float temp_time;
	timespec h_start, h_end;
	clock_gettime(CLOCK_REALTIME, &h_start);

	//Lock mutex to check condition variable. Proceed just if the neighbor 
	//wrote the border during the previous iteration
	pthread_mutex_lock(&mutex[neighbor]);
		while(wAcc[neighbor] != step+1)
		{
			pthread_cond_wait(&wCond[neighbor], &mutex[neighbor]);
		}
	pthread_mutex_unlock(&mutex[neighbor]);

	clock_gettime(CLOCK_REALTIME, &h_end);
	sync_time[device] += getTime(h_start, h_end);
	cudaEventRecord(start, 0);	

	cudaMemcpy2D(dev_addr, dev_pitch, host_addr, host_pitch, width_byte, height, cudaMemcpyHostToDevice);

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop); 
	
	cudaEventElapsedTime(&temp_time, start, stop);
	cpy_time[device] += temp_time;

	clock_gettime(CLOCK_REALTIME, &h_start);

	//Lock mutex to increase the read counter.
	pthread_mutex_lock(&mutex[neighbor]);
		rAcc[neighbor]++;
	pthread_mutex_unlock(&mutex[neighbor]);
	pthread_cond_signal(&rCond[neighbor]);

	clock_gettime(CLOCK_REALTIME, &h_end);
	sync_time[device] += getTime(h_start, h_end);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);
}

/*
 * tetra_thread
 *	params - contains the parameters required to setup a partial solver (i.e. the solver running on one of the GPU and solving 
 *	computing a specific portion of domain).
 */

void * tetra_thread(void *params)
{
	//Cast the cookie pointer to the right type
	struct Params *p = (struct Params *)params;

	struct timespec h_start, h_end;

	clock_gettime(CLOCK_REALTIME, &h_start);
	
	selectDevice(p->device);

	clock_gettime(CLOCK_REALTIME, &h_end);

	printf("device selection time : %.9fs\n", dev_sel_time[p->device] = getTime(h_start, h_end));

	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	
	float r_kernel_time = 0, b_kernel_time = 0, temp_time;
	
	//Setup subdomain processing
	float *dgrid;
	int offset, height;
	size_t pitch;
	
	//Total number of rows counting borders. GPUs in the middle
	//deal with two sides.
	if((p->device == 0) || (p->device == p->threads-1))
		height = p->h+(p->border_size);
	else
		height = p->h+(2*p->border_size);
	
	//Offset to the beginning of the subdomain
	if(p->device == 0)
		offset = 0;
	else
		offset = p->border_size*p->n;

	cudaEventRecord(start, 0);	

	//Allocate memory on the device
	cudaMallocPitch((void **)&dgrid, &pitch, p->n*sizeof(float), height);

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop); //Block until the event is actually recorded
	cudaEventElapsedTime(&temp_time, start, stop);
	printf("t%d allocation on device in %f ms.\n", p->device, temp_time);

	printf("t%d pitch: %ld, height: %d, offset: %d.\n", p->device, pitch, height, offset);
	printf("t%d grid: %p, dgrid: %p.\n", p->device, p->grid, dgrid);
	
	//Kernel Execution Configuration
	int grid_x, grid_y, local_w;

	//Number of threads per threadblock
	local_w = 112;
	//Number of threadblocks per grid. Strip partitioning, the
	//horizontal dimension doesn't change.
	//Blocks shift of (local_w-2) elements horizontally.
	grid_x = p->n/(local_w-2);	
	if(grid_x*(local_w-2)<p->n) grid_x++;
	
	printf("t%d local_w: %d\n", p->device, local_w);
	printf("t%d grid_x: %d\n", p->device, grid_x);
				
	dim3 blockDim(local_w);

	//First transfer. Entire subdomain+borders.
	cudaEventRecord(start, 0);	

	cudaMemcpy2D(dgrid, pitch, p->grid-offset, p->n*sizeof(float), p->n*sizeof(float), height, cudaMemcpyHostToDevice);

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop); 
	cudaEventElapsedTime(&temp_time, start, stop);
	printf("t%d copied first %d rows from %p to %p in %f ms.\n", p->device, height, p->grid-offset, dgrid, temp_time);
	ext_cpy_time[p->device] = temp_time;
	cpy_time[p->device] = 0;
	comm_time[p->device] = 0;

	clock_gettime(CLOCK_REALTIME, &h_start);
	
	//Increase read counters.
	if(p->device == 0)
	{
		pthread_mutex_lock(&mutex[1]);
			rAcc[1]++;
		pthread_mutex_unlock(&mutex[1]);
		pthread_cond_signal(&rCond[1]);
	}
	else if(p->device == p->threads-1)
	{
		pthread_mutex_lock(&mutex[2*p->device-2]);
			rAcc[2*p->device-2]++;
		pthread_mutex_unlock(&mutex[2*p->device-2]);
		pthread_cond_signal(&rCond[2*p->device-2]);
	}
	else
	{
		pthread_mutex_lock(&mutex[2*p->device-2]);
			rAcc[2*p->device-2]++;
		pthread_mutex_unlock(&mutex[2*p->device-2]);
		pthread_cond_signal(&rCond[2*p->device-2]);

		pthread_mutex_lock(&mutex[2*p->device+1]);
			rAcc[2*p->device+1]++;
		pthread_mutex_unlock(&mutex[2*p->device+1]);
		pthread_cond_signal(&rCond[2*p->device+1]);
	}
	clock_gettime(CLOCK_REALTIME, &h_end);
	ext_sync_time[p->device] = getTime(h_start, h_end);
	sync_time[p->device] = 0;

// CORE COMPUTATION - Begin //////////////////////////

	//So to be able to iterate with border_size == 0
	int b = p->border_size > 0 ? p->border_size : 1;
	
	for(int i = 0; i < p->iterations/b; i++)
	{	
		int border = b;
		//k is an offset used to point to the beginning of the area to process
		int k=0;
		while(border > 0)
		{
			//Compute the vertical dimension of the grid.
			//Blocks shift with window size 2.
			grid_y = (2*height-4)/sizeof(float) + (height)%2; 
			dim3 gridDim(grid_x, grid_y);
			
			switch(p->solver)
			{
				case 0: 
				{	
					/* apply R/B SOR, using global mem for fetching data. */					
					cudaEventRecord(start, 0);	
					
					//Red kernel
					rsor_gm<<<gridDim, blockDim, 4*local_w*sizeof(float)>>>(dgrid+k*(pitch/sizeof(float)), p->n, height, p->omega, pitch/sizeof(float), 
						(p->h*p->device)-(p->device>0)*border, p->device);

					cudaEventRecord(stop, 0);
					cudaEventSynchronize(stop); 
					cudaEventElapsedTime(&temp_time, start, stop);
					r_kernel_time += temp_time;

					cudaEventRecord(start, 0);	

					//Black kernel
					bsor_gm<<<gridDim, blockDim, 4*local_w*sizeof(float)>>>(dgrid+k*(pitch/sizeof(float)), p->n, height, p->omega, pitch/sizeof(float), 
						(p->h*p->device)-(p->device>0)*border, p->device); 

					cudaEventRecord(stop, 0);
					cudaEventSynchronize(stop); //Block until the event is actually recorded
					cudaEventElapsedTime(&temp_time, start, stop);
					b_kernel_time += temp_time;

					break;
				}
				case 1: 
				{	
					/* apply R/B SOR, using a texture to fetch data */
					size_t tex_offset = selectTextureToBind(p->device, i, dgrid+k*(pitch/sizeof(float)), height*pitch);
					
					cudaEventRecord(start, 0);	

					//Red kernel
					rsor_lintex<<<gridDim, blockDim, 4*local_w*sizeof(float)>>>(dgrid+k*(pitch/sizeof(float)), p->n, height, p->omega, pitch/sizeof(float), 
						(p->h*p->device)-(p->device>0)*border, tex_offset/sizeof(float), p->device);

					cudaEventRecord(stop, 0);
					cudaEventSynchronize(stop); 
					cudaEventElapsedTime(&temp_time, start, stop);
					r_kernel_time += temp_time;

					cudaEventRecord(start, 0);	

					//Black kernel
					bsor_lintex<<<gridDim, blockDim, 4*local_w*sizeof(float)>>>(dgrid+k*(pitch/sizeof(float)), p->n, height, p->omega, pitch/sizeof(float), 
						(p->h*p->device)-(p->device>0)*border, tex_offset/sizeof(float), p->device); 

					cudaEventRecord(stop, 0);
					cudaEventSynchronize(stop);
					cudaEventElapsedTime(&temp_time, start, stop);
					b_kernel_time += temp_time;

					break;
				}
			}
			
			border--; 
			//increase k everywhere but on device 0
			k += 1*(p->device > 0); 
			//Decrease on outer devices and subtract 2 inside
			height -= 1 + ((p->device > 0) && (p->device < p->threads-1)); 
		} //End border consuming while
		
		//Restore height
		height += b*(1 + ((p->device) && (p->device < p->threads-1)));

//***** Communication *****
		clock_gettime(CLOCK_REALTIME, &h_start);

		if(p->threads == 1)
		{
			//when using one device just continue
			wAcc[p->device]++;
			
			clock_gettime(CLOCK_REALTIME, &h_end);
			comm_time[p->device] += getTime(h_start, h_end);

			continue;
		}
		
		//Transfer using chessboard ordering. 0<->1 2<->3 .. (n-2)<->(n-1) -----> 1<->2 ... (n-3)<->(n-2)
		if((p->device == 0) || (p->device == p->threads-1))
		{
			switch(p->device)
			{
				case 0:
				{
					write_ghosts(0, 0, 1, i, p->grid+(p->h-p->border_size)*p->n, p->n*sizeof(float), dgrid+(p->h-p->border_size)*(pitch/sizeof(float)), pitch, p->n*sizeof(float), 
						p->border_size);
					read_ghosts(0, 0, 1, i, p->grid+(p->h*p->n), p->n*sizeof(float), dgrid+(p->h)*(pitch/sizeof(float)), pitch, p->n*sizeof(float), p->border_size);
					break;
				}
				default:
				{
					write_ghosts(p->device, 2*p->device-1, 2*p->device-2, i, p->grid, p->n*sizeof(float), dgrid+(p->border_size)*(pitch/sizeof(float)), pitch, p->n*sizeof(float), 
						p->border_size);
					read_ghosts(p->device, 2*p->device-1, 2*p->device-2, i, p->grid-offset, p->n*sizeof(float), dgrid, pitch, p->n*sizeof(float), p->border_size);
				}
			}
		} else
		{
			//if odd start from north otherwise from south
			if(p->device&1) 
			{
				write_ghosts(p->device, p->device*2 - 1, p->device*2 - 2, i, p->grid, p->n*sizeof(float), dgrid+(p->border_size)*(pitch/sizeof(float)), pitch, p->n*sizeof(float),
					p->border_size);
				read_ghosts(p->device, p->device*2 - 1, p->device*2 - 2, i, p->grid-(p->border_size*p->n), p->n*sizeof(float), dgrid, pitch, p->n*sizeof(float), 
					p->border_size);

				write_ghosts(p->device, p->device*2, p->device*2 + 1, i, p->grid+(p->h-p->border_size)*p->n, p->n*sizeof(float), 
					dgrid+(p->h)*(pitch/sizeof(float)), pitch, p->n*sizeof(float), p->border_size);
				read_ghosts(p->device, p->device*2, p->device*2 + 1, i, p->grid+(p->h*p->n), p->n*sizeof(float), dgrid+((p->h)+(p->border_size))*(pitch/sizeof(float)), 
					pitch, p->n*sizeof(float), p->border_size);
			} else
			{
				write_ghosts(p->device, p->device*2, p->device*2 + 1, i, p->grid+(p->h-p->border_size)*p->n, p->n*sizeof(float), 
					dgrid+(p->h)*(pitch/sizeof(float)), pitch, p->n*sizeof(float), p->border_size);
				read_ghosts(p->device, p->device*2, p->device*2 + 1, i, p->grid+(p->h*p->n), p->n*sizeof(float), dgrid+((p->h)+(p->border_size))*(pitch/sizeof(float)), 
					pitch, p->n*sizeof(float), p->border_size);

				write_ghosts(p->device, p->device*2 - 1, p->device*2 - 2, i, p->grid, p->n*sizeof(float), dgrid+(p->border_size)*(pitch/sizeof(float)), pitch, p->n*sizeof(float),
					p->border_size);
				read_ghosts(p->device, p->device*2 - 1, p->device*2 - 2, i, p->grid-(p->border_size*p->n), p->n*sizeof(float), dgrid, pitch, p->n*sizeof(float), 
					p->border_size);
			}
		}

		clock_gettime(CLOCK_REALTIME, &h_end);
		comm_time[p->device] += getTime(h_start, h_end);

//***** End Communication *****
	}

// CORE COMPUTATION - End //////////////////////////
	
	//Last transfer. Just subdomain
	if(p->device == 0)
	{
		cudaEventRecord(start, 0);

		cudaMemcpy2D(p->grid, p->n*sizeof(float), dgrid, pitch, p->n*sizeof(float), p->h, cudaMemcpyDeviceToHost);

		cudaEventRecord(stop, 0);
		cudaEventSynchronize(stop);
		cudaEventElapsedTime(&temp_time, start, stop);
		ext_cpy_time[p->device] += temp_time;
		printf("t%d final copy from dev(%p) to host(%p) in %f ms.\n", p->device, dgrid, p->grid, temp_time);
	} else
	{
		cudaEventRecord(start, 0);

		cudaMemcpy2D(p->grid, p->n*sizeof(float), dgrid+p->border_size*(pitch/sizeof(float)), pitch, p->n*sizeof(float), p->h, cudaMemcpyDeviceToHost);

		cudaEventRecord(stop, 0);
		cudaEventSynchronize(stop);
		cudaEventElapsedTime(&temp_time, start, stop);
		ext_cpy_time[p->device] += temp_time;
		printf("t%d final copy from dev(%p) to host(%p) in %f ms.\n", p->device, dgrid, p->grid, temp_time);
	}

	//Format and display output
	if(p->threads==1)
		b = p->iterations;
	printf("t%d avarage RSOR kernel time per iteration: %f ms.\n", p->device, r_kernel_time/p->iterations);
	printf("t%d avarage BSOR kernel time per iteration: %f ms.\n", p->device, b_kernel_time/p->iterations);
	printf("t%d total kernel time: %f ms.\n", p->device, r_kernel_time+b_kernel_time);
	printf("t%d MKT: %f ms.\n", p->device, (r_kernel_time+b_kernel_time)/p->iterations);
	
	printf("t%d total transfer time: %f ms [iter: %f + ext: %f].\n", p->device, cpy_time[p->device]+ext_cpy_time[p->device], 
		cpy_time[p->device], ext_cpy_time[p->device]);
	printf("t%d MTT: %f ms.\n", p->device, cpy_time[p->device]/(p->iterations/b));
	printf("t%d total synchronization time: %f ms [iter: %f + ext: %f].\n", p->device, (sync_time[p->device]+ext_sync_time[p->device])*1E3, 
		sync_time[p->device]*1E3, ext_sync_time[p->device]*1E3);
	printf("t%d MST: %f ms.\n", p->device, (sync_time[p->device]*1E3)/(p->iterations/b));
	printf("t%d total communication time: %f ms [iter: %f + ext: %f].\n", p->device, 
		comm_time[p->device]*1E3+ext_sync_time[p->device]*1E3+ext_cpy_time[p->device], comm_time[p->device]*1E3, 
		ext_sync_time[p->device]*1E3+ext_cpy_time[p->device]);
	printf("t%d MCT: %f ms.\n", p->device, (comm_time[p->device]*1E3)/(p->iterations/b));

	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	cudaFree(dgrid);	
	
	return NULL;
}

/*
 * call_solver
 * 	solver - the index of the solver to use.
 * 	grid - pointer to the PDE domain.
 *		n - width/height of the domain in terms of number of elements per row/column.
 *		iterations - the number of times the solver should iterate.
 *		threads - indicates how many threads, and finally devices, should cooperate in solving the PDE.
 *
 * Instantiates the right number of threads and manages their cooperation to solve the PDE.
 */

int call_solver(int solver, float *omega, float *grid, int m, int n, int iterations, int threads, int border_size)
{
	pthread_t *t_id = new pthread_t[threads];
	
	struct Params *t_p = new struct Params[threads];

	struct timespec start, end;
	
	mutex = new pthread_mutex_t[2*(threads-1)];
	rCond = new pthread_cond_t[2*(threads-1)];
	wCond = new pthread_cond_t[2*(threads-1)];
	rAcc = new int[2*(threads-1)];
	wAcc = new int[2*(threads-1)];
	
	ext_cpy_time = new double[threads];
	ext_sync_time = new double[threads];
	cpy_time = new double[threads];
	sync_time = new double[threads];
	comm_time = new double[threads];
	dev_sel_time = new double[threads];

	printf("border_size: %d.\n", border_size);

	if(*omega == 0)
	{
		float pi = 4.0*atan(1.0);
		printf("pi: %.15f\n", pi);
		float interval = 1.0/(n+1);
		*omega = 2.0/(1.0+sin(pi*interval));
		if(*omega >1.9) *omega = 1.9;
	}

	printf("omega: %f.\n", *omega);
	
	for(int i=0; i<2*(threads-1); i++)
	{
		pthread_mutex_init(&mutex[i], NULL); 
		pthread_cond_init(&rCond[i], NULL); 
		pthread_cond_init(&wCond[i], NULL); 
		rAcc[i] = 0;
		wAcc[i] = 0;
	}

	for(int i=0; i<threads; i++)
	{
		t_p[i].solver = solver;
		t_p[i].omega = *omega;
		t_p[i].n = n;
		t_p[i].h = m/threads; //supposes threads%2 = 0
		t_p[i].border_size = border_size*(threads>1);
		t_p[i].iterations = iterations;
		t_p[i].grid = grid + i*t_p[i].h*n;
		t_p[i].device = i;
		t_p[i].threads = threads;
	}
	
	if(t_p[0].h*threads < m) 
		t_p[threads-1].h += m-(t_p[0].h*threads);

	clock_gettime(CLOCK_REALTIME, &start);
	
	for(int i=0; i<threads; i++)
		pthread_create(&t_id[i], NULL, &tetra_thread, &t_p[i]);

	for(int i=0; i<threads; i++)
	{
		pthread_join(t_id[i], NULL);
		printf("t%d joined.\n", i);		
	}

	clock_gettime(CLOCK_REALTIME, &end);
	
	double max_dev_sel = getMaxDeviceSelectionTime(threads);
	printf("Elapsed time: %.9fs\n", getTime(start, end)-max_dev_sel);

	for(int i=0; i<2*(threads-1); i++)
	{
		pthread_mutex_destroy(&mutex[i]);
		pthread_cond_destroy(&rCond[i]);
		pthread_cond_destroy(&wCond[i]);
	}

	delete [] mutex;
	delete [] rCond;
	delete [] wCond;
	delete [] rAcc;
	delete [] wAcc;

	delete [] t_id;
	delete [] t_p;

	delete [] ext_cpy_time;
	delete [] ext_sync_time;
	delete [] comm_time;
	delete [] sync_time;
	delete [] cpy_time;
	delete [] dev_sel_time;
	
	return 0;
}
