/**
 * Tetra is a parallel PDE Solver for Dirichlet Problems. 
 * The PDE is solved using a multi-GPU supported, 5-points stencil, Red-Black SOR.
 **/
 
#include <cstdlib>
#include <sstream>
#include <getopt.h>

#include <tetra.h>

//command-line options (short version)
const char * const short_options = "ht:m:n:f:i:b:k:o:s";

//command-line options (long version)
const struct option long_options[] = {
	{"help",				0,	NULL,	'h'},
	{"threads",			1,	NULL,	't'},
	{"rows",				1,	NULL,	'm'},
	{"columns",			1,	NULL,	'n'},
	{"file",				1,	NULL,	'f'},
	{"iterations",		1,	NULL, 'i'},
	{"border_size",	1,	NULL, 'b'},
	{"kernel",			1,	NULL, 'k'},
	{"omega",			1,	NULL, 'o'},
	{"save_image",		0,	NULL,	's'}
};

/*
 * Print usage information to <stream> and exit with <exit_code>.
 */
void print_usage(ostream* stream, string name, int exit_code)
{
	*stream << "Usage: " << name << " -t <threads> -m <rows> -n <columns> -f <file> -i <iterations> -b <border size> -k <kernel> -o <omega> [-s] [-h]" << endl;
	
	exit(exit_code);
}

/*
 * MAIN FUNCTION
 */

int main(int argc, char * argv[])
{
	int threads, m, n, iterations, bs, kernel, save_img = 0;
	float omega;
	string filename;
		
	int next_opt, opt_counter = 0;
	
	do{
		next_opt = getopt_long(argc, argv, short_options, long_options, NULL);
		
		switch(next_opt)
		{
			case 'h':
				print_usage(&cout, argv[0], 0);
			case 't':
				stringstream(optarg) >> threads;
				opt_counter++;
				break;
			case 'm':
				stringstream(optarg) >> m;
				opt_counter++;
				break;
			case 'n':
				stringstream(optarg) >> n;
				opt_counter++;
				break;
			case 'f':
				filename = optarg;
				opt_counter++;
				break;
			case 'i':
				stringstream(optarg) >> iterations;
				opt_counter++;
				break;
			case 'b':
				stringstream(optarg) >> bs;
				opt_counter++;
				break;
			case 'k':
				stringstream(optarg) >> kernel;
				opt_counter++;
				break;
			case 'o':
				stringstream(optarg) >> omega;
				opt_counter++;
				break;
			case 's':
				save_img=1;
				break;
			case '?':
				cerr << "Unexpected option." << endl;
				print_usage(&cerr, argv[0], 1);
			case -1:
				//Done with options
				break;
			default:
				exit(-1);
		}
	} while(next_opt != -1);
	
	if(opt_counter != 8)
	{
		cerr << "Missing mandatory options." << endl;
		print_usage(&cerr, argv[0], 1);
	}
	
	if(bs > iterations)
	{
		cerr << "Border_size can't be greater than iterations. Abort." << endl;
		exit(2);
	}
	
	//Instianciate the solver
	Tetra t(filename, m, n);
	
	//Solve the problem
	t.solve(kernel, &omega, bs, iterations, threads);
	
	//Eventually save output image
	stringstream ss;
	ss << "gout/k" << kernel << "s" << m << "x" << n << ".t" << threads << ".b" << bs << ".i" << iterations << ".o" <<  omega << ".tiff";
	if(save_img)
	{
		cout << "Image..." << t.saveImage(ss.str().c_str()) << endl;
		cout << "format: kernel.size.#threads.border_size.iterations.omega.tiff" << endl;
	}
	
	exit(0);
}
