/**
 * Tetra PDE solver's member functions implementation.
 **/

#include <tiffio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdlib.h>

#include <tetra.h>
#include <errmng.h>
#include <errno.h>

///////////////// CUDA wrappers //////////////////////////////////
extern "C"
	int call_solver(int solver, float *omega, float *grid, int m, int n, int iterations, int threads, int ghosts);
//////////////////////////////////////////////////////////////////

/*
 * Tetra Constructor.
 */

Tetra::Tetra(string ifilename, int m, int n)
{
	if((fd = open(ifilename.c_str(), O_RDWR)) == -1)
	{
		cerr << "Error: problems while opening input file." << endl;
		exit(-1);
	}
	struct stat fs;

	fstat(fd, &fs);

	cout << "Mapping " << fs.st_size << " bytes to memory..." << endl;
	if((data = (float *)mmap(0, (size_t)m*n*sizeof(float), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_POPULATE, fd, 0)) == MAP_FAILED)
	{
		cerr << "Error: problems while mapping input file." << endl;
		perror(NULL);
		exit(-1);
	}

	cout << "Done." << endl;

	dim_y = m;
	dim_x = n;
}

/*
 * Tetra Destructor
 */

Tetra::~Tetra()
{
	munmap((void *)data, (size_t)dim_x*dim_y*sizeof(float));
	close(fd);
}

/*
 * Tetra::solve
 * 	solver - indicates the index of the solver to use.
 * 	iterations - indicates the number of iterations to use in solving the PDE.
 */

int Tetra::solve(int solver, float *omega, int border_size, int iterations, int threads)
{
	call_solver(solver, omega, data, dim_y, dim_x, iterations, threads, border_size);	
}

/*
 * Tetra::saveImage
 * 	title - image title
 *
 * Save data in TIFF format (greyscale, 1 sample/pixel, 8 bit/pixel).
 * The hotter the whiter.
 */

int Tetra::saveImage(string title)
{
	int result = 0;
	unsigned char *raster = new unsigned char[dim_y*dim_x];

	for (int y=0; y<dim_y; y++)
		for (int x=0; x<dim_x; x++)
			raster[y*dim_x+x] = (unsigned char)(data[y*dim_x+x]/MAX_TEMP * 255);

	TIFF* tif = TIFFOpen(title.c_str(), "w");

	if(tif)
	{
		TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, dim_x);
		TIFFSetField(tif, TIFFTAG_IMAGELENGTH, dim_y);
		TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 8);
		TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1);
		TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, 1);
		TIFFSetField(tif, TIFFTAG_PLANARCONFIG, 1);

		for (int r = 0; r < dim_y; r++)
			if(TIFFWriteScanline(tif, &raster[r*dim_x], r, 1) != 1)
			{
				result = -1;
				break;
			}
	} else result = -1;

	TIFFClose ( tif );
	
	delete [] raster;
	    
	return result;
}

/*
 * Tetra::printArea
 *
 * Print 2D data to screen.
 *
 *            (x_s,y_s)-------(x_e,y_s)
 *                |               |
 *            (x_s,y_e)-------(x_e,y_e)
 *
 */
void Tetra::printArea(int x_s, int x_e, int y_s, int y_e)
{
	for(int j=y_s; j<y_e; j++)
	{
		for(int i=x_s; i<x_e; i++)
		{
			cout << data[j*dim_x+i] << " ";
		}
		cout << endl;
	}
}

/*
 * Tetra::printGrid
 *
 * Print 2D the entire domain.
 */

void Tetra::printGrid()
{
	printArea(0, dim_x, 0, dim_y);
}
