#include <pde_kernels.cu>

#include <stdio.h>
#include <time.h>

void checkCUDAError(const char *msg)
{
	cudaError_t err = cudaGetLastError();
	if(cudaSuccess != err)
	{
		fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString(err));
		//exit(EXIT_FAILURE);
	}
}

double getTime(timespec start, timespec end)
{
	long nsec = end.tv_nsec-start.tv_nsec;
	if(nsec < 0)
	{
		nsec = 1E9+nsec;
		end.tv_sec-=1;
	}
	return (double)(end.tv_sec-start.tv_sec)+(double)(nsec*1E-9);
}

// Code that gives very accurate timer.
static unsigned long long rdtsctime()
{
    unsigned int eax, edx;
    unsigned long long val;
    __asm__ __volatile__("rdtsc":"=a"(eax), "=d"(edx));
    val = edx;
    val = val << 32;
    val += eax;
    return val;
}

//Empty kernel 

__global__ void test()
{
//...void...
}

int main()
{
	cudaEvent_t start, stop;
	float time;
	
	double *foo0 = new double[4];

	for(int i = 0; i< 4; i++)
	{
		foo0[i] = 0;
	}
	
	for(int i = 0; i< 10000; i++)
	{
		foo0[0] += i; 
		foo0[1] += i; 
		foo0[2] += i; 
		foo0[3] += i; 
	}

	delete [] foo0;
		
	timespec clock_start, clock_end;
	long long cycle_start, cycle_end;
	
	cycle_start = rdtsctime();
	cycle_end = rdtsctime();
	cycle_end = rdtsctime();
	
	printf("Time with rdtsctime: %f us.\n", (cycle_end-cycle_start)/(3.2*1E9)*1E6);

	clock_gettime(CLOCK_REALTIME, &clock_start);
	cycle_end = rdtsctime();
	clock_gettime(CLOCK_REALTIME, &clock_end);
	
	printf("Time with clock_gettime: %f us.\n", getTime(clock_start, clock_end)*1E6);
		
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	
	size_t pitch;
	
	float *hgrid = (float *)malloc(11585*11585*sizeof(float));
	
	float *dgrid;
	
	cudaEventRecord(start, 0);
	cudaMallocPitch((void **)&dgrid, &pitch, 11585*sizeof(float), 11585);
	checkCUDAError(NULL);
	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
		
	cudaEventElapsedTime(&time, start, stop);
	printf("CudaMallocPitch 2 GB: %fms\n", time);
	printf("Pitch: %ld\n", pitch);
	
	float startup;
	cudaEventRecord(start, 0);
	cudaMemcpy2D(dgrid, pitch, hgrid, 11585*sizeof(float), 0, 0, cudaMemcpyHostToDevice);
	checkCUDAError(NULL);
	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
		
	cudaEventElapsedTime(&startup, start, stop);
	printf("CudaMemcpy2D startup: %fms\n", startup);
	
	float mean = 0.f;
	for(int i=0; i<1000; i++)
	{
		cudaEventRecord(start, 0);
		cudaMemcpy2D(dgrid, pitch, hgrid, 11585*sizeof(float), 1, 1, cudaMemcpyHostToDevice);
		checkCUDAError(NULL);
		cudaEventRecord(stop, 0);
		cudaEventSynchronize(stop);
		
		cudaEventElapsedTime(&time, start, stop);
		mean += time;
	}
	float tword = mean/1000-startup;
	printf("CudaMemcpy2D 1 Byte: %fms\n", tword);

	cudaEventRecord(start, 0);
	test<<<1, 1>>>();
	checkCUDAError(NULL);
	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
		
	cudaEventElapsedTime(&time, start, stop);
	printf("Empty kernel: %fus\n", time*1000.f);

	float temp;
	cudaEventRecord(start, 0);
	cudaMemcpy2D(dgrid, pitch, hgrid, 11585*sizeof(float), 11585*sizeof(float), 11585, cudaMemcpyHostToDevice);
	checkCUDAError(NULL);
	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
		
	cudaEventElapsedTime(&temp, start, stop);
	printf("CudaMemcpy2D transferring prediction for 2 GB: %fms\n", 11585.f*(startup+11585.f*sizeof(float)*tword));
	printf("CudaMemcpy2D to send 11585x11585: %fms\n", temp);
	
	cudaEventRecord(start, 0);
	cudaMemcpy2D(dgrid, pitch, hgrid, 11585*sizeof(float), 5792*sizeof(float), 5, cudaMemcpyHostToDevice);
	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
		
	cudaEventElapsedTime(&temp, start, stop);
	printf("CudaMemcpy2D transferring prediction for kn/2: %fms\n", 5*(startup+5792*sizeof(float)*tword));
	printf("CudaMemcpy2D to send kn/2: %fms\n", temp);
	
	cudaEventRecord(start, 0);
	cudaMemcpy2D(dgrid, pitch, hgrid, 11585*sizeof(float), 5*sizeof(float), 5792, cudaMemcpyHostToDevice);
	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
		
	cudaEventElapsedTime(&temp, start, stop);
	printf("CudaMemcpy2D transferring prediction for n/2*k: %fms\n", 5792*(startup+5*sizeof(float)*tword));
	printf("CudaMemcpy2D to send n/2*k: %fms\n", temp);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	return 0;
}
